import os
import unittest
import datetime as dt
from pathlib import Path
import re
from .pytest_common import create_black_oil_simulator
from opm.io.parser import Parser
from opm.io.ecl_state import EclipseState
from opm.io.schedule import Schedule
from opm.io.summary import SummaryConfig
from .pytest_common import pushd

class TestBasic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_dir = Path(os.path.dirname(__file__))
        cls.data_dir = test_dir.parent.joinpath("test_data/SPE1CASE1b")


    def test_all(self):
        with pushd(self.data_dir):
            self.deck  = Parser().parse('SPE1CASE1.DATA')
            state = EclipseState(self.deck)
            self.schedule = Schedule( self.deck, state )
            summary_config = SummaryConfig(self.deck, state, self.schedule)
            self.unit_system = self.deck.active_unit_system()

            self.assertTrue('PROD' in self.schedule)
            self.assertTrue('INJ'  in self.schedule)
            self.assertEqual(dt.datetime(2015, 1, 1),   self.schedule.start)
            self.assertEqual(dt.datetime(2016, 1, 1), self.schedule.end)
            self.sim = create_black_oil_simulator(
                self.deck, state, self.schedule, summary_config)
            tsteps = self.schedule.reportsteps
            self.assertEqual(dt.datetime(2015, 1, 1), tsteps[0])
            last_step = len(tsteps) - 1
            self.assertEqual(dt.datetime(2016, 1, 1), tsteps[last_step])
            self.sim.step_init()
            report_step = 4
            self.sim.advance(report_step=report_step)
            well_name = "PROD"
            prod = self.schedule.get_well(well_name, 2)
            self.assertEqual(prod.status(), "OPEN")
            #schedule.shut_well("PROD", 3)
            #prod = schedule.get_well("PROD", 3)
            #self.assertEqual(prod.status(), "SHUT")
            self.subtest_modify_prod_weltarg_dynamically(well_name, report_step)
            self.sim.step()
            report_step = self.sim.current_step()
            well_name = "INJ"
            self.subtest_modify_inj_weltarg_dynamically(well_name, report_step)
            self.sim.advance(report_step=last_step)
            self.sim.step_cleanup()

    def subtest_modify_inj_weltarg_dynamically(self, well_name, report_step):
        prop = self.schedule.get_injection_properties(well_name, report_step)
        self.assertEqual(prop['surf_inj_rate'], 100000.0) # Mscf/day
        self.assertEqual(prop['resv_inj_rate'], 0.0) # rb/day
        self.assertEqual(prop['bhp_target'], 9014.0) # psi
        self.assertEqual(prop['thp_target'], 0.0)
        new_grat_target = prop['surf_inj_rate'] - 100  # stb/day
        self.update_inj_grat_target_wconinje(well_name, new_grat_target)
        self.sim.step()
        prop2 = self.schedule.get_injection_properties(well_name, report_step+1)
        self.assertEqual(prop2['surf_inj_rate'], new_grat_target)
        new_grat_target += 200
        self.update_inj_grat_target_weltarg(well_name, new_grat_target)
        self.sim.step()
        prop3 = self.schedule.get_injection_properties(well_name, report_step+2)
        self.assertEqual(prop3['surf_inj_rate'], new_grat_target)

    def subtest_modify_prod_weltarg_dynamically(self, well_name, report_step):
        prop = self.schedule.get_production_properties(well_name, report_step)
        self.assertEqual(prop['alq_value'], 0.0)
        self.assertEqual(prop['bhp_target'], 1000.0)
        self.assertEqual(prop['gas_rate'], 0.0)
        self.assertEqual(prop['liquid_rate'], 0.0)
        self.assertEqual(prop['oil_rate'], 20000.0)
        self.assertEqual(prop['resv_rate'], 0.0)
        self.assertEqual(prop['thp_target'], 0.0)
        self.assertEqual(prop['water_rate'], 0.0)
        new_oil_target = prop['oil_rate'] + 10000  # stb/day
        self.update_prod_orat_target_wconprod(well_name, new_oil_target)
        self.sim.step()
        prop2 = self.schedule.get_production_properties(well_name, report_step+1)
        self.assertEqual(prop2['oil_rate'], new_oil_target)
        new_oil_target += 1000
        self.update_prod_orat_target_weltarg(well_name, new_oil_target)
        self.sim.step()
        prop3 = self.schedule.get_production_properties(well_name, report_step+2)
        self.assertEqual(prop3['oil_rate'], new_oil_target)

    # This is an alternative to using WELTARG
    def update_inj_grat_target_wconinje(self, well_name, new_surf_flow_rate):
        data = self.deck["WCONINJE"]
        # assumes data looks like this:
        #  WCONINJE
        #  	'INJ'	'GAS'	'OPEN'	'RATE'	100000 1* 9014 /
        #  /
        # The initial rate can also be obtained from data[0][4].get_uda(0).get_double()
        data = re.sub(pattern='100000', repl=str(new_surf_flow_rate), string=str(data), count=1)
        report_step = self.sim.current_step()
        self.schedule.insert_keywords(
            data, step=report_step, unit_system=self.unit_system)

    # This is an alternative to using WCONINJE to modify injection properties
    def update_inj_grat_target_weltarg(self, well_name, net_surf_flow_rate):
        data = """
WELTARG
    '{}'  GRAT {} /
/
        """.format(well_name, net_surf_flow_rate)
        report_step = self.sim.current_step()
        self.schedule.insert_keywords(
            data, step=report_step, unit_system=self.unit_system)


    # This is an alternative to using WCONPROD to modify production properties
    def update_prod_orat_target_weltarg(self, well_name, oil_target):
        data = """
WELTARG
    '{}'  ORAT {} /
/
        """.format(well_name, oil_target)
        report_step = self.sim.current_step()
        self.schedule.insert_keywords(
            data, step=report_step, unit_system=self.unit_system)

    # This is an alternative to using WELTARG to modify production properties
    def update_prod_orat_target_wconprod(self, well_name, oil_target):
        well_status = "OPEN"
        control_mode = "ORAT"
        bhp_limit = 1000  # psia
        data = """
WCONPROD
	'{}' '{}' '{}' {} 4* {} /
/
        """.format(well_name, well_status, control_mode, oil_target, bhp_limit)
        report_step = self.sim.current_step()
        self.schedule.insert_keywords(
            data, step=report_step, unit_system=self.unit_system)



import os
import unittest
from pathlib import Path
import numpy as np
import subprocess

from utils import Dense, Sequential
from utils import collect_input_features, compute_output_vars, write_config, extract_newtit_from_file, extract_unrst_variables
from utils import ABSOLUTE_CASES, RELATIVE_CASES, FEATURE_ENGINEERING_CASES, SCALING_CASES, MULTI_MODEL_CASES, ZERO_NEWTON_CASES, ALL_CASES

from opm.ml.ml_tools.kerasify import export_model

class TestHybridNewton(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Run baseline simulation via subprocess and collect timesteps & fluid state."""
        test_dir = Path(__file__).parent
        cls.data_dir = test_dir.parent.parent.parent / "test_data/SPE1CASE1b"
        cls.deck_file = "SPE1CASE1.DATA"

        cls.unrst = {}
        cls.rd_init = {"PERMX": []}
        cls.times = []

        baseline_cmd = [
            os.environ.get("FLOW_BINARY", default="flow"),
            str(cls.data_dir / cls.deck_file),
            "--output-extra-convergence-info=steps,iterations",
            "--newton-min-iterations=0",
            "--output-dir=.",
            "--full-time-step-initially=true",
        ]

        print("Running baseline simulation... ")
        subprocess.run(baseline_cmd, stdout=subprocess.DEVNULL, check=True)

        # Extract Newton iterations
        cls.baseline_iters = extract_newtit_from_file(Path('.') / Path(cls.deck_file).stem)
        print(f"Baseline Newton iterations: {cls.baseline_iters}")

        # Extract fluid state from UNRST for all timesteps
        cls.unrst, cls.times = extract_unrst_variables(".", cls.deck_file)
        cls.rd_init["PERMX"] = [10**2] * len(cls.unrst["PRESSURE"][0])
        cls.n_cells = len(cls.unrst["PRESSURE"][0])

    def _run_cases(self, cases):
        models_dir = Path('.') / "models"
        os.makedirs(models_dir, exist_ok=True)

        for case in cases:
            with self.subTest(case=case["label"]):
                json_path = self._build_case_models(case, models_dir)
                self.assertTrue(json_path and os.path.isfile(json_path),
                                f"Config JSON was not created for case {case['label']}")

                # Run hybrid Newton test
                hybrid_iters = self._run_hybrid_newton_test(json_path)
        return hybrid_iters

    def _build_case_models(self, case, models_dir):
        models_to_write = []

        if case.get("mode", "single") == "single":
            model_entries = [case]
            multi_mode = False
        else:
            model_entries = case.get("models", [])
            multi_mode = True

        for model_case in model_entries:
            start_idx = model_case.get("start_idx", case.get("start_idx"))
            end_idx = model_case.get("end_idx", case.get("end_idx"))
            input_features = model_case["input_features"]
            output_features = model_case["output_features"]
            fe_input = model_case.get("feature_engineering_input")
            fe_output = model_case.get("feature_engineering_output")
            scaling_input = model_case.get("scaling_input")
            scaling_output = model_case.get("scaling_output")

            input_flat, input_scaling_params = collect_input_features(
                input_features, start_idx, self.unrst, self.rd_init, self.times,
                feat_eng=fe_input, scaling=scaling_input
            )
            output_flat, output_scaling_params = compute_output_vars(
                output_features, start_idx, end_idx, self.unrst, self.rd_init,
                feature_engineering=fe_output, scaling=scaling_output,
            )

            input_flat = input_flat.reshape(-1)
            output_flat = output_flat.reshape(-1)

            model = Sequential([
                Dense(input_dim=input_flat.shape[0], output_dim=output_flat.shape[0])
            ])
            model.layers[0].set_weights([np.zeros((input_flat.shape[0], output_flat.shape[0])), output_flat])
            apply_time = self.times[start_idx]

            active_cells = list(range(self.n_cells))

            models_dir.mkdir(exist_ok=True)
            model_name = "_".join(input_features) + "__TO__" + "_".join(output_features) + f"_{start_idx}_{end_idx}.model"
            model_path = models_dir / model_name
            export_model(model, model_path)

            models_to_write.append({
                "input_features": input_features,
                "output_features": output_features,
                "model_path": model_path,
                "active_cells": active_cells,
                "apply_time": apply_time,
                "feature_engineering_input": fe_input,
                "feature_engineering_output": fe_output,
                "scaling_input": scaling_input,
                "scaling_output": scaling_output,
                "input_scaling_params": input_scaling_params,
                "output_scaling_params": output_scaling_params
            })

        if not models_to_write:
            print(f"No valid models for {case['label']}")
            return None

        return write_config(models_to_write, models_dir, multi_mode=multi_mode, case_label=case['label'])

    def _run_hybrid_newton_test(self, json_path):
        """Run hybrid Newton via subprocess and extract Newton iterations."""
        cmd = [
            os.environ.get("FLOW_BINARY", default="flow"),
            str(self.data_dir / self.deck_file),
            f"--hybrid-newton-config-file={json_path}",
            "--use-hybrid-newton=true",
            "--output-extra-convergence-info=steps,iterations",
            "--newton-min-iterations=0",
            "--output-dir=.",
            "--full-time-step-initially=true"
        ]

        print(f"Running hybrid Newton simulation with config {json_path}...")
        subprocess.run(cmd,
                       stdout=subprocess.DEVNULL,
                       check=True)

        hybrid_iters = extract_newtit_from_file(Path('.') / Path(self.deck_file).stem)
        return hybrid_iters

    def test_absolute_cases(self):
        self._run_cases(ABSOLUTE_CASES)

    def test_relative_cases(self):
        self._run_cases(RELATIVE_CASES)

    def test_feature_engineering_cases(self):
        self._run_cases(FEATURE_ENGINEERING_CASES)

    def test_scaling_cases(self):
        self._run_cases(SCALING_CASES)

    def test_multi_model_cases(self):
        self._run_cases(MULTI_MODEL_CASES)

    def test_zero_newton_cases(self):
        hybrid_iters = self._run_cases(ZERO_NEWTON_CASES)
        print(f"Hybrid Newton iterations all_features: {hybrid_iters}")

    def test_all_cases(self):
        hybrid_iters = self._run_cases(ALL_CASES)
        print(f"Hybrid Newton iterations all_cases: {hybrid_iters}")


if __name__ == "__main__":
    unittest.main()



import logging
import zipfile

from pathlib import Path

import click

from opm_wheels import click_helpers, helpers
from opm_wheels.click_helpers import ClickOptions
from opm_wheels.constants import Directories, PythonVersion
from opm_wheels.config_reader import get_default_versions

@click.group()
@click.option("--verbose", "-v", is_flag=True, help="Show verbose output")
@click.pass_context
def main(ctx: click.Context, verbose: bool) -> None:
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
        # logging.basicConfig(level=logging.WARNING)


@main.command()
@click.argument("name", type=str)
@ClickOptions.wheel_dir
def show_wheel_files(name: str, wheel_dir: str) -> None:
    """Show the files in wheel file. NAME is the name of a wheel file, relative to the
    python/wheelhouse directory."""
    if not name.endswith(".whl"):
        raise click.BadParameter("NAME must end with .whl")
    wheel_file = helpers.get_wheel_file(wheel_dir, name)
    # Check that
    with zipfile.ZipFile(wheel_file, "r") as zf:
        for name in zf.namelist():
            print(name)

@main.command()
@click.option(
    "--docker-os", "-d",
    type=str,
    default="ubuntu-22.04",
    help=(
        "The OS to use in the Docker image. Supported (canonical): "
        "ubuntu-22.04, ubuntu-24.04, debian-11. Legacy colon form accepted."
    )
)
@click.option(
    "--opm-simulators-repo", "-osr",
    type=str,
    default="https://github.com/OPM/opm-simulators.git",
    help="The URL of the opm-simulators repository."
)
@click.option(
    "--opm-common-repo", "-ocr",
    type=str,
    default="https://github.com/OPM/opm-common.git",
    help="The URL of the opm-common repository."
)
@click.option(
    "--opm-simulators-branch", "-osb",
    type=str,
    default="master",
    help="The branch of the opm-simulators repository."
)
@click.option(
    "--opm-common-branch", "-ocb",
    type=str,
    default="master",
    help="The branch of the opm-common repository."
)
@click.option(
    "--docker-progress", "-dp",
    is_flag=True,
    default=False,
    help="Show Docker progress output."
)
@click.option(
    "--python-versions", "-p",
    type=str,
    callback=click_helpers.validate_python_versions,
    help="Python versions to install in the Docker image. Valid values are: "
            f"{', '.join(PythonVersion.valid_versions())}. If not specified, all supported versions are installed."
)
def build_docker_image(
    docker_os: str,
    opm_simulators_repo: str,
    opm_common_repo: str,
    opm_simulators_branch: str,
    opm_common_branch: str,
    docker_progress: bool,
    python_versions: list[PythonVersion] = None
) -> None:
    """Build the Docker image. The option "--docker-os specifies the name of the OS to use
    in the Docker image. Supported: ubuntu:22.04, ubuntu:24.04, debian:11. If option --docker-progress
    is given, the Docker build will use option "--progress=plain". The default value is False."""

    testing_root = helpers.get_testing_root_dir()
    # Use docker directory as build context so python_versions.json is accessible
    docker_root = testing_root.parent  # python/docker/test_wheels -> python/docker
    canonical_os = helpers.canonicalize_docker_os(docker_os)
    docker_tag = helpers.get_docker_tag(canonical_os)
    # Prefer canonical directory, fall back to legacy colon directory (pre-rename)
    dockerfile = testing_root / Directories.docker_files / canonical_os / "Dockerfile"
    if not dockerfile.exists():
        legacy_os = canonical_os.replace("-", ":", 1)
        candidate = testing_root / Directories.docker_files / legacy_os / "Dockerfile"
        if candidate.exists():
            dockerfile = candidate
        else:
            raise FileNotFoundError(f"Dockerfile {dockerfile} does not exist")
    # Build the Docker image
    helpers.run_docker_build(
        docker_tag,
        opm_simulators_repo,
        opm_common_repo,
        opm_simulators_branch,
        opm_common_branch,
        build_ctxt_dir=docker_root,
        docker_file=dockerfile,
        docker_progress=docker_progress,
        python_versions=python_versions
    )

@main.command()
@click.option(
    "--docker-os", "-d",
    type=str,
    default="ubuntu-22.04",
    help=(
        "The OS for the Docker image. Supported (canonical): "
        "ubuntu-22.04, ubuntu-24.04, debian-11. Legacy colon form accepted."
    )
)
@ClickOptions.wheel_dir
@click.option(
    "--python-versions", "-p",
    type=str,
    callback=click_helpers.validate_python_versions,
    help="The Python versions to use for testing. Valid values are: "
            f"{', '.join(PythonVersion.valid_versions())}"
)
@click.option(
    "--host-tests-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    help="Directory containing opm-common and opm-simulators repositories from host. "
         "If specified, tests will be run from host directories instead of cloned repos. "
         "Expected structure: host-tests-dir/opm-common/python and host-tests-dir/opm-simulators/python"
)
@click.option(
    "--test-cases-common",
    type=str,
    help="Comma-separated list of specific test case patterns to run for opm-common (e.g., 'eclfile,esmry'). "
         "Will run test_<pattern>.py files. If not specified, all tests are run."
)
@click.option(
    "--test-cases-simulators",
    type=str,
    help="Comma-separated list of specific test case patterns to run for opm-simulators (e.g., 'basic,mpi'). "
         "Will run test_<pattern>.py files. If not specified, all tests are run."
)
@click.option(
    "--stop-on-error",
    is_flag=True,
    default=False,
    help="Stop execution on first test failure. If not specified, all tests will run even if some fail."
)
def run_docker_image(
    docker_os: str,
    wheel_dir: str,
    python_versions: list[PythonVersion] | None,
    host_tests_dir: Path = None,
    test_cases_common: str = None,
    test_cases_simulators: str = None,
    stop_on_error: bool = False,
) -> None:
    f"""Run the Docker image. The option "--docker-tag" specifies the tag of the Docker
    image to run. The option "--python-versions" specifies the Python versions to
    use for testing. Valid values are: {', '.join(PythonVersion.valid_versions())}. The value should
    be a comma-separated list of versions without separating space. For example: 3.8,3.9,3.10.
    If not given, the default value is {','.join(get_default_versions())}."""
    wheel_dir = helpers.get_wheel_abs_dir(wheel_dir)
    canonical_os = helpers.canonicalize_docker_os(docker_os)
    docker_tag = helpers.get_docker_tag(canonical_os)
    helpers.run_docker_run(docker_tag, wheel_dir, python_versions, host_tests_dir, test_cases_common, test_cases_simulators, stop_on_error)



import os
import unittest
from pathlib import Path
from .pytest_common import pushd, create_black_oil_simulator, create_gas_water_simulator, create_onephase_simulator

class TestBasic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_dir = Path(os.path.dirname(__file__))
        cls.data_dir_bo = test_dir.parent.joinpath("test_data/SPE1CASE1a")
        cls.data_dir_op = test_dir.parent.joinpath("test_data/SPE1CASE1")
        cls.data_dir_gw = test_dir.parent.joinpath("test_data/SPE1CASE2")

    # IMPORTANT: Since all the python unittests run in the same process we must be
    #  careful to not call MPI_Init() more than once.
    #  Tests are run alphabetically, so we need to make sure that
    #  the the first test calls MPI_Init(), therefore the name of the tests
    #  have a numeric label like "01" in test_01_bo to ensure that they
    #  are run in a given order.

    # IMPORTANT:This test must be run first since it calls MPI_Init()
    def test_01_blackoil(self):
        with pushd(self.data_dir_bo):
            sim = create_black_oil_simulator(args=['--output-dir=01_fluidstate'], filename="SPE1CASE1.DATA")
            sim.setup_mpi(True, False)
            sim.step_init()
            sim.step()
            oil_pressure = sim.get_fluidstate_variable(name='po')
            self.assertAlmostEqual(oil_pressure[0], 35795160.67, delta=1e4, msg='value of oil pressure')
            gas_pressure = sim.get_fluidstate_variable(name='pg')
            self.assertAlmostEqual(gas_pressure[0], 35795160.67, delta=1e4, msg='value of gas pressure')
            water_pressure = sim.get_fluidstate_variable(name='pw')
            self.assertAlmostEqual(water_pressure[0], 35795160.67, delta=1e4, msg='value of water pressure')
            rho_w = sim.get_fluidstate_variable(name='rho_w')
            self.assertAlmostEqual(rho_w[0], 998.9822355, places=3, msg='value of water density')
            rho_g = sim.get_fluidstate_variable(name='rho_g')
            self.assertAlmostEqual(rho_g[0], 241.36955087, places=2, msg='value of gas density')
            rho_o = sim.get_fluidstate_variable(name='rho_o')
            self.assertAlmostEqual(rho_o[0], 631.78674, places=2, msg='value of oil density')
            Rs = sim.get_fluidstate_variable(name='Rs')
            self.assertAlmostEqual(Rs[0], 226.19666048, places=5, msg='value of solution gas-oil ratio')
            Rv = sim.get_fluidstate_variable(name='Rv')
            self.assertAlmostEqual(Rv[0], 0.0, places=5, msg='value of volatile gas-oil ratio')
            Sw = sim.get_fluidstate_variable(name='Sw')
            self.assertAlmostEqual(Sw[0], 0.11969486712, places=5, msg='value of water saturation')
            So = sim.get_fluidstate_variable(name='So')
            self.assertAlmostEqual(So[0], 0.825129, places=3, msg='value of oil saturation')
            Sg = sim.get_fluidstate_variable(name='Sg')
            self.assertAlmostEqual(Sg[0], 0.055138968544, places=3, msg='value of gas saturation')
            T = sim.get_fluidstate_variable(name='T')
            self.assertAlmostEqual(T[0], 288.705, places=3, msg='value of temperature')

    def test_02_onephase(self):
        with pushd(self.data_dir_op):
            sim = create_onephase_simulator(args=["--output-dir=02_fluidstate"], filename="SPE1CASE1_WATER.DATA")
            sim.setup_mpi(False, False)
            sim.step_init()
            sim.step()
            water_pressure = sim.get_fluidstate_variable(name='pw')
            self.assertAlmostEqual(water_pressure[0], 44780102.277570, delta=1e4, msg='value of water pressure')
            rho_w = sim.get_fluidstate_variable(name='rho_w')
            self.assertAlmostEqual(rho_w[0], 1003.182858, places=3, msg='value of water density')
            Sw = sim.get_fluidstate_variable(name='Sw')
            self.assertAlmostEqual(Sw[0], 1.0, places=5, msg='value of water saturation')

    # IMPORTANT: This test must be run last since it calls MPI_Finalize()
    def test_99_gaswater(self):
        with pushd(self.data_dir_gw):
            sim = create_gas_water_simulator(args=["--output-dir=99_fluidstate"], filename="SPE1CASE2_GASWATER.DATA")
            sim.setup_mpi(False, True)
            sim.step_init()
            sim.step()
            gas_pressure = sim.get_fluidstate_variable(name='pg')
            self.assertAlmostEqual(gas_pressure[0], 32410549.874824, delta=1e4, msg='value of gas pressure')
            water_pressure = sim.get_fluidstate_variable(name='pw')
            self.assertAlmostEqual(water_pressure[0], 32410549.874824, delta=1e4, msg='value of water pressure')
            rho_g = sim.get_fluidstate_variable(name='rho_g')
            self.assertAlmostEqual(rho_g[0], 219.613424, delta=1e-2, msg='value of gas density')
            rho_w = sim.get_fluidstate_variable(name='rho_w')
            self.assertAlmostEqual(rho_w[0], 1047.849235, delta=1e-3, msg='value of water density')
            Sg = sim.get_fluidstate_variable(name='Sg')
            self.assertAlmostEqual(Sg[0], 0.708648, delta=1e-2, msg='value of gas saturation')
            Sw = sim.get_fluidstate_variable(name='Sw')
            self.assertAlmostEqual(Sw[0], 0.291351, delta=1e-2, msg='value of water saturation')



# Instead of having the pybind11 extension module, e.g.
#   simulators.cpython-310-x86_64-linux-gnu.so  located
#   directly in the opm directory, we create a package (sub
#   directory) with the same name and place it there.
#   In this way we can do (if needed in the future)
#
#  from opm.simulators import BlackOilSimulator, FoamSimulator, PurePythonUtils, ...
#
#  where FoamSimulator and PurePythonUtils does not currently exists,
#  but could be possible future extensions..
#
from .BlackOil import BlackOilSimulator
from .GasWater import GasWaterSimulator
from .OnePhase import OnePhaseSimulator



import os
import unittest
from pathlib import Path
from .pytest_common import pushd, create_black_oil_simulator, create_gas_water_simulator, create_onephase_simulator

class TestBasic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # NOTE: See comment in test_basic.py for the reason why we are
        #   only using a single test_all() function instead of splitting
        #   it up in multiple test functions
        test_dir = Path(os.path.dirname(__file__))
        cls.data_dir_bo = test_dir.parent.joinpath("test_data/SPE1CASE1a")
        cls.data_dir_op = test_dir.parent.joinpath("test_data/SPE1CASE1")
        cls.data_dir_gw = test_dir.parent.joinpath("test_data/SPE1CASE2")

    # IMPORTANT: Since all the python unittests run in the same process we must be
    #  careful to not call MPI_Init() more than once.
    #  Tests are run alphabetically, so we need to make sure that
    #  the the first test calls MPI_Init(), therefore the name of the tests
    #  have a numeric label like "01" in test_01_bo to ensure that they
    #  are run in a given order.

    # IMPORTANT:This test must be run first since it calls MPI_Init()
    def test_01_blackoil(self):
        with pushd(self.data_dir_bo):
            sim = create_black_oil_simulator(args=["--output-dir=01_primary"], filename="SPE1CASE1.DATA")
            sim.setup_mpi(True, False)
            sim.step_init()
            sim.step()
            pressure = sim.get_primary_variable(variable='pressure')
            self.assertAlmostEqual(pressure[0], 35795160.67, delta=1e4, msg='value of pressure')
            pressure_meaning = sim.get_primary_variable_meaning(
                variable='pressure')
            pressure_meaning_map = sim.get_primary_variable_meaning_map(
                variable='pressure')
            self.assertEqual(pressure_meaning[0], pressure_meaning_map["Po"])
            water_meaning = sim.get_primary_variable_meaning(
                variable='water')
            water_meaning_map = sim.get_primary_variable_meaning_map(
                variable='water')
            self.assertEqual(water_meaning[0], water_meaning_map["Sw"])
            gas_meaning = sim.get_primary_variable_meaning(
                variable='gas')
            gas_meaning_map = sim.get_primary_variable_meaning_map(
                variable='gas')
            self.assertEqual(gas_meaning[0], gas_meaning_map["Sg"])
            brine_meaning = sim.get_primary_variable_meaning(
                variable='brine')
            brine_meaning_map = sim.get_primary_variable_meaning_map(
                variable='brine')
            self.assertEqual(brine_meaning[0], brine_meaning_map["Disabled"])

    def test_02_onephase(self):
        with pushd(self.data_dir_op):
            sim = create_onephase_simulator(args=["--output-dir=02_primary"], filename="SPE1CASE1_WATER.DATA")
            sim.setup_mpi(False, False)
            sim.step_init()
            sim.step()
            pressure = sim.get_primary_variable(variable='pressure')
            self.assertAlmostEqual(pressure[0], 44780102.277570, delta=1e4, msg='value of pressure')
            pressure_meaning = sim.get_primary_variable_meaning(
                variable='pressure')
            pressure_meaning_map = sim.get_primary_variable_meaning_map(
                variable='pressure')
            self.assertEqual(pressure_meaning[0], pressure_meaning_map["Pw"])
            water_meaning = sim.get_primary_variable_meaning(
                variable='water')
            water_meaning_map = sim.get_primary_variable_meaning_map(
                variable='water')
            self.assertEqual(water_meaning[0], water_meaning_map["Disabled"])
            brine_meaning = sim.get_primary_variable_meaning(
                variable='brine')
            brine_meaning_map = sim.get_primary_variable_meaning_map(
                variable='brine')
            self.assertEqual(brine_meaning[0], brine_meaning_map["Disabled"])

    # IMPORTANT: This test must be run last since it calls MPI_Finalize()
    def test_99_gaswater(self):
        with pushd(self.data_dir_gw):
            sim = create_gas_water_simulator(args=["--output-dir=99_primary"], filename="SPE1CASE2_GASWATER.DATA")
            sim.setup_mpi(False, True)
            sim.step_init()
            sim.step()
            pressure = sim.get_primary_variable(variable='pressure')
            self.assertAlmostEqual(pressure[0], 32410549.874824, delta=1e4, msg='value of pressure')
            pressure_meaning = sim.get_primary_variable_meaning(
                variable='pressure')
            pressure_meaning_map = sim.get_primary_variable_meaning_map(
                variable='pressure')
            self.assertEqual(pressure_meaning[0], pressure_meaning_map["Pg"])
            water_meaning = sim.get_primary_variable_meaning(
                variable='water')
            water_meaning_map = sim.get_primary_variable_meaning_map(
                variable='water')
            self.assertEqual(water_meaning[0], water_meaning_map["Sw"])
            gas_meaning = sim.get_primary_variable_meaning(
                variable='gas')
            gas_meaning_map = sim.get_primary_variable_meaning_map(
                variable='gas')
            self.assertEqual(gas_meaning[0], gas_meaning_map["Disabled"])
            brine_meaning = sim.get_primary_variable_meaning(
                variable='brine')
            brine_meaning_map = sim.get_primary_variable_meaning_map(
                variable='brine')
            self.assertEqual(brine_meaning[0], brine_meaning_map["Disabled"])



#! /usr/bin/env python3

# This script is called as ENTRYPOINT from the Dockerfiles, e.g. docker/test_wheels/dockerfiles/.../Dockerfile
# For each python version in PYTHON_VERSIONS (with the values in docker/python_versions.json as default), it:
#  - installs the wheels for the python version
#  - runs the opm-common and opm-simulators python unit tests
# Returns 0 if all tests passed, 1 otherwise

import glob
import json
import os
import shutil
import subprocess
import sys

def should_stop_on_error() -> bool:
    """Check if we should stop on test errors based on environment variable."""
    return os.environ.get("STOP_ON_ERROR") == "1"

def copy_essential_directories(source_dir: str, dest_dir: str, essential_items: list) -> None:
    """Copy essential directories from source to destination, avoiding problematic items.

    Ensures clean slate by removing any existing destination directory first.
    """
    # Clean slate approach - delete and recreate for predictable, reproducible behavior
    if os.path.exists(dest_dir):
        shutil.rmtree(dest_dir)
    os.makedirs(dest_dir)

    for item in essential_items:
        item_path = os.path.join(source_dir, item)
        dest_path = os.path.join(dest_dir, item)

        if os.path.exists(item_path):
            if os.path.isdir(item_path):
                shutil.copytree(item_path, dest_path)  # No dirs_exist_ok needed - directory is guaranteed empty
            else:
                shutil.copy2(item_path, dest_path)

def run_opm_common_tests(python_bin: str) -> bool:
    """Run all unittests in opm-common/python/tests"""
    # Check for specific test selection
    test_cases = os.environ.get("TEST_CASES_COMMON")
    stop_on_error = should_stop_on_error()

    if os.environ.get("USE_HOST_TESTS"):
        # Copy strategy for host-mounted directories:
        # When using --host-tests-dir, the host directories contain 'opm/' source code
        # that conflicts with installed wheel packages. The local 'opm/' directories
        # shadow the installed packages, causing import failures because the local
        # source code doesn't have compiled extensions that the wheels provide.
        # Solution: Copy only test files (not conflicting opm/ directories) to /tmp
        # and run tests from there, allowing imports to find the installed wheel packages.
        test_dir = "/tmp/opm-common-tests"

        # Copy essential directories to clean location, avoiding conflicting opm/ directory
        copy_essential_directories("/test/opm-common/python", test_dir, ["tests", "test", "examples"])

        # Run tests from clean location (preserves package structure for relative imports)
        if test_cases:
            # Run specific test files
            all_passed = True
            for test_case in test_cases.split(","):
                test_case = test_case.strip()
                print(f"Running opm-common test: {test_case}")
                result = subprocess.run([
                    python_bin, "-m", "unittest", f"tests.test_{test_case}",
                    "-v"
                ], cwd=test_dir, check=False)
                if result.returncode != 0:
                    print(f"Test {test_case} failed with return code {result.returncode}", file=sys.stderr)
                    all_passed = False
                    if stop_on_error:
                        raise subprocess.CalledProcessError(result.returncode, result.args)
            if not all_passed:
                print("Some opm-common specific tests failed", file=sys.stderr)
            return all_passed
        else:
            # Run all tests
            result = subprocess.run([
                python_bin, "-m", "unittest", "discover",
                "-s", test_dir,
                "-p", "test_*.py",
                "-v"
            ], cwd=test_dir, check=False)
            if result.returncode != 0:
                print(f"opm-common tests failed with return code {result.returncode}", file=sys.stderr)
                if stop_on_error:
                    raise subprocess.CalledProcessError(result.returncode, result.args)
                return False
            return True
    else:
        if test_cases:
            # Run specific test files in cloned repo environment
            all_passed = True
            for test_case in test_cases.split(","):
                test_case = test_case.strip()
                print(f"Running opm-common test: {test_case}")
                result = subprocess.run([
                    python_bin, "-m", "unittest", f"test_{test_case}",
                    "-v"
                ], cwd="/test/opm-common/python", check=False)
                if result.returncode != 0:
                    print(f"Test {test_case} failed with return code {result.returncode}", file=sys.stderr)
                    all_passed = False
                    if stop_on_error:
                        raise subprocess.CalledProcessError(result.returncode, result.args)
            if not all_passed:
                print("Some opm-common specific tests failed", file=sys.stderr)
            return all_passed
        else:
            # Run from cloned repos (opm directory was moved to opm_original during build)
            result = subprocess.run([
                python_bin, "-m", "unittest", "discover",
                "-s", "/test/opm-common/python",  # path to the tests
                "-v"  # verbose output
            ], check=False)
            if result.returncode != 0:
                print(f"opm-common tests failed with return code {result.returncode}", file=sys.stderr)
                if stop_on_error:
                    raise subprocess.CalledProcessError(result.returncode, result.args)
                return False
            return True

def run_opm_simulators_tests(python_bin: str) -> bool:
    """Run all unittests in opm-simulators/python/test"""
    # Check for specific test selection
    test_cases = os.environ.get("TEST_CASES_SIMULATORS")
    stop_on_error = should_stop_on_error()

    if os.environ.get("USE_HOST_TESTS"):
        # Copy strategy for host-mounted directories:
        # Same issue as opm-common - host 'opm/' directories conflict with installed wheels.
        # Copy entire python directory structure except conflicting opm/ directory
        test_dir = "/tmp/opm-simulators-tests"

        # Copy essential directories to clean location, avoiding conflicting opm/ directory
        copy_essential_directories("/test/opm-simulators/python", test_dir, ["test", "test_data"])

        # Run tests from clean location (preserves package structure for relative imports)
        # Use parent directory as start and specify test pattern to make relative imports work
        if test_cases:
            # Run specific test files
            all_passed = True
            for test_case in test_cases.split(","):
                test_case = test_case.strip()
                print(f"Running opm-simulators test: {test_case}")
                result = subprocess.run([
                    python_bin, "-m", "unittest", f"test.test_{test_case}",
                    "-v"
                ], cwd=test_dir, check=False)
                if result.returncode != 0:
                    print(f"Test {test_case} failed with return code {result.returncode}", file=sys.stderr)
                    all_passed = False
                    if stop_on_error:
                        raise subprocess.CalledProcessError(result.returncode, result.args)
            if not all_passed:
                print("Some opm-simulators specific tests failed", file=sys.stderr)
            return all_passed
        else:
            # Run all tests
            result = subprocess.run([
                python_bin, "-m", "unittest", "discover",
                "-s", test_dir,
                "-p", "test_*.py",
                "-v"
            ], cwd=test_dir, check=False)
            if result.returncode != 0:
                print(f"opm-simulators tests failed with return code {result.returncode}", file=sys.stderr)
                if stop_on_error:
                    raise subprocess.CalledProcessError(result.returncode, result.args)
                return False
            return True
    else:
        if test_cases:
            # Run specific test files in cloned repo environment
            all_passed = True
            for test_case in test_cases.split(","):
                test_case = test_case.strip()
                print(f"Running opm-simulators test: {test_case}")
                result = subprocess.run([
                    python_bin, "-m", "unittest", f"test.test_{test_case}",
                    "-v"
                ], cwd="/test/opm-simulators/python", check=False)
                if result.returncode != 0:
                    print(f"Test {test_case} failed with return code {result.returncode}", file=sys.stderr)
                    all_passed = False
                    if stop_on_error:
                        raise subprocess.CalledProcessError(result.returncode, result.args)
            if not all_passed:
                print("Some opm-simulators specific tests failed", file=sys.stderr)
            return all_passed
        else:
            # Run from cloned repos (opm directory was moved to opm_original during build)
            result = subprocess.run([
                python_bin, "-m", "unittest", "discover",
                "-s", "/test/opm-simulators/python",  # path to the tests
                "-v"  # verbose output
            ], check=False)
            if result.returncode != 0:
                print(f"opm-simulators tests failed with return code {result.returncode}", file=sys.stderr)
                if stop_on_error:
                    raise subprocess.CalledProcessError(result.returncode, result.args)
                return False
            return True


def install_and_test(python_version, pyenv_full_version):
    """Install wheels for a single Python version and run a basic import test.
       Return True if everything succeeded, or False if it failed."""

    # Determine the correct wheel tags
    if tuple(map(int, python_version.split('.'))) < (3, 8):
        py_tag = f"cp{python_version.replace('.', '')}-cp{python_version.replace('.', '')}m"
    else:
        py_tag = f"cp{python_version.replace('.', '')}-cp{python_version.replace('.', '')}"

    # Find wheel files dynamically instead of hardcoding version
    wheel_path = f"/test/wheels"
    simulators_wheels = glob.glob(f"{wheel_path}/opm_simulators-*-{py_tag}-*manylinux*x86_64.whl")
    common_wheels = glob.glob(f"{wheel_path}/opm-*-{py_tag}-*manylinux*x86_64.whl")

    if not simulators_wheels:
        raise FileNotFoundError(f"No opm_simulators wheel found for Python {python_version} in {wheel_path}")
    if not common_wheels:
        raise FileNotFoundError(f"No opm-common wheel found for Python {python_version} in {wheel_path}")

    wheel_file_simulators = simulators_wheels[0]  # Use first match
    wheel_file_common = common_wheels[0]  # Use first match

    # Absolute path to the specific Python version
    python_bin = f"/root/.pyenv/versions/{pyenv_full_version}/bin/python"

    print(f"\n--- Installing wheels and running tests with Python {python_version} ---")
    print(f"Found opm-common wheel: {os.path.basename(wheel_file_common)}")
    print(f"Found opm_simulators wheel: {os.path.basename(wheel_file_simulators)}")

    try:
        # 1) Upgrade pip
        subprocess.run(
            [python_bin, "-m", "pip", "install", "--upgrade", "pip"],
            check=True
        )

        # 2) Install the wheels
        subprocess.run(
            [python_bin, "-m", "pip", "install",
             wheel_file_common,
             wheel_file_simulators],
            check=True
        )

        # 3) Run unittests based on what's requested
        test_cases_common = os.environ.get("TEST_CASES_COMMON")
        test_cases_simulators = os.environ.get("TEST_CASES_SIMULATORS")

        # Track overall test success
        overall_success = True

        # If no specific test cases are specified, run all tests
        if not test_cases_common and not test_cases_simulators:
            simulators_success = run_opm_simulators_tests(python_bin)
            common_success = run_opm_common_tests(python_bin)
            overall_success = simulators_success and common_success
        else:
            # Run only the requested test suites
            if test_cases_simulators or not test_cases_common:
                # Run simulators tests if explicitly requested OR if common tests not specified
                simulators_success = run_opm_simulators_tests(python_bin)
                overall_success = overall_success and simulators_success
            if test_cases_common or not test_cases_simulators:
                # Run common tests if explicitly requested OR if simulators tests not specified
                common_success = run_opm_common_tests(python_bin)
                overall_success = overall_success and common_success

        # Report overall result
        if overall_success:
            return True
        elif should_stop_on_error():
            # When stop_on_error is True, failures should stop execution
            return False
        else:
            # When stop_on_error is False, continue even if tests failed
            print(f"Tests failed for Python {python_version}, but continuing due to continue-on-error mode", file=sys.stderr)
            return True

    except subprocess.CalledProcessError:
        print(f"Installation or test failed for Python {python_version}", file=sys.stderr)
        return False

def main():
    # Try to load version map, with fallback handling
    version_map_file = "/test/common/python_version_map.json"

    try:
        with open(version_map_file) as f:
            version_map = json.load(f)
        print(f"Loaded version map from {version_map_file}")
    except FileNotFoundError:
        print(f"Error: Version map file not found at {version_map_file}")
        print("This file should be created during Docker image build by install_python_versions.py")
        sys.exit(1)
    except json.JSONDecodeError as e:
        print(f"Error: Invalid JSON in {version_map_file}: {e}")
        sys.exit(1)

    # Auto-detect actually installed Python versions instead of using all from JSON
    # This allows runtime to use only the versions that were installed during build
    try:
        result = subprocess.run(["pyenv", "versions", "--bare"],
                              capture_output=True, text=True, check=True)
        installed_versions = []
        for line in result.stdout.strip().split('\n'):
            if line.strip():
                # Extract short version (e.g., "3.12.9" -> "3.12")
                full_version = line.strip()
                short_version = '.'.join(full_version.split('.')[:2])
                if short_version in version_map:
                    installed_versions.append(short_version)

        if installed_versions:
            default_versions = ",".join(sorted(set(installed_versions)))
            print(f"Auto-detected installed Python versions: {default_versions}")
        else:
            # Fallback to JSON versions if detection fails
            default_versions = ",".join(version_map.keys())
            print(f"Could not detect installed versions, using all from JSON: {default_versions}")
    except Exception as e:
        # Fallback to JSON versions if detection fails
        default_versions = ",".join(version_map.keys())
        print(f"Could not detect installed versions ({e}), using all from JSON: {default_versions}")

    python_versions_env = os.environ.get("PYTHON_VERSIONS", default_versions)

    print(f"Running tests.. Requested Python versions: {python_versions_env}")

    python_versions = [v.strip() for v in python_versions_env.split(",") if v.strip()]

    # Validate requested versions against available versions
    invalid_versions = []
    for ver in python_versions:
        if ver not in version_map:
            invalid_versions.append(ver)

    if invalid_versions:
        print(f"Error: Invalid Python version(s) requested: {', '.join(invalid_versions)}")
        print(f"   Available versions: {', '.join(sorted(version_map.keys()))}")
        sys.exit(1)

    # Track how many pass/fail
    results = []
    for ver in python_versions:
        full_version = version_map[ver]
        passed = install_and_test(ver, full_version)
        results.append((ver, passed))

    # Summarize
    total_tests = len(results)
    passed_tests = sum(1 for _, passed in results if passed)
    failed_tests = total_tests - passed_tests

    print("\n--- Summary of tests ---")
    if failed_tests > 0:
        print(f"{failed_tests}/{total_tests} tests failed.")
        print("result: FAILED")
        sys.exit(1)
    else:
        print(f"{passed_tests}/{total_tests} tests passed.")
        print("result: OK")
        sys.exit(0)

if __name__ == "__main__":
    main()



from opm_wheels.constants import PythonVersion
from typing import Union
import click

from opm_wheels import helpers

def validate_python_versions(
    ctx: click.Context, param: Union[click.Option, click.Parameter], value: str
) -> list[PythonVersion] | None:
    if value is None:
        # Return None to allow runtime auto-detection of installed versions
        return None
    versions = value.split(",")
    versions = [PythonVersion.from_str(str(v)) for v in versions]
    valid_versions = [PythonVersion.from_str(v) for v in PythonVersion.values()]
    for version in versions:
        if version not in valid_versions:
            raise click.BadParameter(
                f"Invalid python version '{version}'. Must be one of {', '.join([str(v) for v in valid_versions])}."
            )
    return versions

class ClickOptions:
    wheel_dir = lambda func: click.option(
        "--wheel-dir", "-wd",
        type=str,
        default=str(helpers.get_wheel_default_dir()),
        help=f"The directory containing the wheels. Can be relative to the root of the opm-simulators repository, or an absolute path. The default value is {str(helpers.get_wheel_default_dir())} directory."
    )(func)



import enum

class BaseEnum(enum.Enum):
    """Base class for enums with keys and values."""

    @classmethod
    def keys(cls) -> set[str]:
        return set(cls.__members__.keys())

    @classmethod
    def values(cls) -> set[str]:
        return {member.value for member in cls}


class Directories:
    docker_files = "dockerfiles"
    python = "python"
    wheelhouse = "wheelhouse"


# Load Python versions from JSON to create enum dynamically
def _get_python_versions_for_enum():
    """Get Python versions from JSON for enum creation."""
    try:
        from pathlib import Path
        import json

        # Get path to python_versions.json
        current_dir = Path(__file__).parent
        config_path = current_dir / ".." / ".." / ".." / ".." / "python_versions.json"
        config_path = config_path.resolve()

        with open(config_path) as f:
            config = json.load(f)

        return list(config["supported_versions"].keys())
    except Exception:
        # Fallback to hardcoded values
        return ["3.8", "3.9", "3.10", "3.11", "3.12", "3.13", "3.14"]

# Dynamically create PythonVersion enum using functional API
_python_versions = _get_python_versions_for_enum()

# Create enum members dict
enum_members = {}
for version in _python_versions:
    enum_name = f'v{version.replace(".", "_")}'
    enum_members[enum_name] = version

# Create PythonVersion enum dynamically
PythonVersion = enum.Enum('PythonVersion', enum_members, type=BaseEnum)

# Add class methods
@classmethod
def _from_str(cls, version_str):
    try:
        return cls[f"v{version_str.replace('.', '_')}"]
    except KeyError:
        valid_versions = ', '.join([member.value for member in cls])
        raise ValueError(f"Invalid python version: {version_str}, valid versions are {valid_versions}")

def _str(self):
    # Return the version in the desired format (e.g., "3.8" instead of "v3_8")
    return self.value

@classmethod
def _valid_versions(cls) -> list[str]:
    """Return a list of valid python versions."""
    return [member.value for member in cls]

# Bind methods to class
PythonVersion.from_str = _from_str
PythonVersion.__str__ = _str
PythonVersion.valid_versions = _valid_versions



import os
import json
import subprocess

# This script is called from the Dockerfiles, e.g. docker/test_wheels/dockerfiles/.../Dockerfile
# It is used to install the Python versions specified in docker/python_versions.json
try:
    # Use the file copied by Docker build
    config_file = "/test/python_versions.json"
    with open(config_file) as f:
        config = json.load(f)
    full_version_map = config["supported_versions"]

    # Check if specific Python versions were requested via build argument
    build_versions = os.environ.get("BUILD_PYTHON_VERSIONS")
    if build_versions:
        requested_versions = [v.strip() for v in build_versions.split(",")]
        version_map = {v: full_version_map[v] for v in requested_versions if v in full_version_map}
        print(f"📋 Using requested Python versions: {list(version_map.keys())}")

        # Validate that all requested versions are supported
        missing_versions = [v for v in requested_versions if v not in full_version_map]
        if missing_versions:
            print(f"Error: Unsupported Python version(s): {missing_versions}")
            print(f"Supported versions: {list(full_version_map.keys())}")
            exit(1)
    else:
        version_map = full_version_map
        print(f"📋 Using all Python versions from python_versions.json: {list(version_map.keys())}")

except FileNotFoundError:
    print("Error: python_versions.json not found!")
    exit(1)
except KeyError as e:
    print(f"Error: Missing key {e} in python_versions.json")
    exit(1)

# Initialize pyenv
subprocess.run('eval "$(pyenv init -)"', shell=True, executable="/bin/bash")

# Install all full versions
for short, full in version_map.items():
    print(f"Installing Python {full} (for {short})...")
    subprocess.check_call(["pyenv", "install", full])

# Set global version to latest
latest_full_version = version_map[max(version_map)]
subprocess.check_call(["pyenv", "global", latest_full_version])
subprocess.run("pyenv rehash", shell=True)

# Create the version map file that entrypoint.py expects
version_map_output = "/test/common/python_version_map.json"
try:
    # Ensure the directory exists
    os.makedirs(os.path.dirname(version_map_output), exist_ok=True)

    # Write the version map to the expected location
    with open(version_map_output, "w") as f:
        json.dump(version_map, f, indent=2)
    print(f"Created version map file at {version_map_output}")
except Exception as e:
    print(f"Warning: Could not create version map file: {e}")

print("Python versions installed and pyenv configured.")



import json
import os
from pathlib import Path
from typing import Dict, List

from opm_wheels.constants import PythonVersion
from opm_wheels.helpers import get_git_root


def get_config_path() -> Path:
    """Get the path to the python_versions.json config file."""
    try:
        # Try to find git root (works when inside repo)
        repo_root = get_git_root()
        return repo_root / "python/docker/python_versions.json"
    except RuntimeError:
        # Fallback to environment variable for global installations
        env_root = os.environ.get('OPM_SIMULATORS_ROOT')
        if env_root:
            repo_root = Path(env_root)
            config_path = repo_root / "python/docker/python_versions.json"
            if config_path.exists():
                return config_path

        raise FileNotFoundError(
            "Could not locate python_versions.json. Either run from within "
            "opm-simulators repository or set OPM_SIMULATORS_ROOT environment variable."
        )


def load_python_config() -> Dict:
    """Load the python versions configuration from JSON."""
    config_path = get_config_path()
    if not config_path.exists():
        raise FileNotFoundError(f"Python config file not found at {config_path}")

    with open(config_path, 'r') as f:
        return json.load(f)


def get_supported_versions() -> Dict[str, str]:
    """Get the mapping of short versions to full versions."""
    config = load_python_config()
    return config["supported_versions"]


def get_default_versions() -> List[str]:
    """Get the list of default Python versions (derived from supported_versions)."""
    config = load_python_config()
    return list(config["supported_versions"].keys())


def get_python_version_objects() -> List[PythonVersion]:
    """Get PythonVersion objects for all default versions."""
    default_versions = get_default_versions()
    return [PythonVersion.from_str(version) for version in default_versions]



import os
import unittest
from pathlib import Path
from .pytest_common import pushd, create_black_oil_simulator, create_gas_water_simulator, create_onephase_simulator

class TestBasic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_dir = Path(os.path.dirname(__file__))
        cls.data_dir_bo = test_dir.parent.joinpath("test_data/SPE1CASE1a")
        cls.data_dir_op = test_dir.parent.joinpath("test_data/SPE1CASE1")
        cls.data_dir_gw = test_dir.parent.joinpath("test_data/SPE1CASE2")

    # IMPORTANT: Since all the python unittests run in the same process we must be
    #  careful to not call MPI_Init() more than once.
    #  Tests are run alphabetically, so we need to make sure that
    #  the the first test calls MPI_Init(), therefore the name of the tests
    #  have a numeric label like "01" in test_01_blackoil to ensure that they
    #  are run in a given order.

    # IMPORTANT: This test must be run first since it calls MPI_Init()
    def test_01_blackoil(self):
        with pushd(self.data_dir_bo):
            sim = create_black_oil_simulator(args=['--linear-solver=ilu0', '--output-dir=01_basic'], filename="SPE1CASE1.DATA")
            sim.setup_mpi(init=True, finalize=False)
            sim.step_init()
            sim.step()
            dt = sim.get_dt()
            # NOTE: the timestep size is reduced to 1 day to avoid regression failures
            # due to changes in time stepping.
            # NOTE: The timestep should be 1 day
            #  = 24 * 60 * 60 seconds = 86400 seconds
            self.assertAlmostEqual(dt, 86400., places=7, msg='value of timestep')
            vol = sim.get_cell_volumes()
            self.assertEqual(len(vol), 300, 'length of volume vector')
            # NOTE: The volume should be 1000 ft x 1000 ft x 20 ft * 0.3 (porosity)
            #  = 600000 ft^3 = 566336.93 m^3
            self.assertAlmostEqual(vol[0], 566336.93, places=2, msg='value of volume')
            poro = sim.get_porosity()
            self.assertEqual(len(poro), 300, 'length of porosity vector')
            self.assertAlmostEqual(poro[0], 0.3, places=7, msg='value of porosity')
            poro = poro *.95
            sim.set_porosity(poro)
            sim.step()
            poro2 = sim.get_porosity()
            self.assertAlmostEqual(poro2[0], 0.285, places=7, msg='value of porosity 2')

    def test_02_onephase(self):
        with pushd(self.data_dir_op):
            sim = create_onephase_simulator(args=['--linear-solver=ilu0', '--output-dir=02_basic'], filename="SPE1CASE1_WATER.DATA")
            sim.setup_mpi(init=False, finalize=False)
            sim.step_init()
            sim.step()
            dt = sim.get_dt()  # 31 days = 31 * 24 * 60 * 60 = 2678400 seconds
            self.assertAlmostEqual(dt, 2678400., places=7, msg='value of timestep')
            vol = sim.get_cell_volumes()
            self.assertEqual(len(vol), 300, 'length of volume vector')
            self.assertAlmostEqual(vol[0], 566336.93, places=2, msg='value of volume')
            poro = sim.get_porosity()
            self.assertEqual(len(poro), 300, 'length of porosity vector')
            self.assertAlmostEqual(poro[0], 0.3, places=7, msg='value of porosity')
            poro = poro *.95
            sim.set_porosity(poro)
            sim.step()
            poro2 = sim.get_porosity()
            self.assertAlmostEqual(poro2[0], 0.285, places=7, msg='value of porosity 2')

    # IMPORTANT: This test must be run last since it calls MPI_Finalize()
    def test_99_gaswater(self):
        with pushd(self.data_dir_gw):
            sim = create_gas_water_simulator(args=['--linear-solver=ilu0', '--output-dir=99_basic'], filename="SPE1CASE2_GASWATER.DATA")
            sim.setup_mpi(init=False, finalize=True)
            sim.step_init()
            sim.step()
            dt = sim.get_dt()  # 31 days = 31 * 24 * 60 * 60 = 2678400 seconds
            self.assertAlmostEqual(dt, 2678400., places=7, msg='value of timestep')
            vol = sim.get_cell_volumes()
            self.assertEqual(len(vol), 300, 'length of volume vector')
            self.assertAlmostEqual(vol[0], 566336.93, places=2, msg='value of volume')
            poro = sim.get_porosity()
            self.assertEqual(len(poro), 300, 'length of porosity vector')
            self.assertAlmostEqual(poro[0], 0.3, places=7, msg='value of porosity')
            poro = poro *.95
            sim.set_porosity(poro)
            sim.step()
            poro2 = sim.get_porosity()
            self.assertAlmostEqual(poro2[0], 0.285, places=7, msg='value of porosity 2')



#!/usr/bin/env python3
"""Helper script to read Python configuration from JSON config file.

This script is called from:
- docker/scripts/generate-pypi-package.sh
- docker/scripts/sync_versions.sh

The script is used to read the Python versions from the docker/python_versions.json config file.
"""

import json
import sys
from pathlib import Path

def get_config_path():
    """Get the path to the python_versions.json config file."""
    script_dir = Path(__file__).parent
    # Config is now at docker root
    config_path = script_dir.parent / "python_versions.json"
    return config_path

def main():
    if len(sys.argv) != 2:
        print("Usage: python read_python_config.py <key>", file=sys.stderr)
        print("Available keys: default_versions (derived), supported_versions", file=sys.stderr)
        sys.exit(1)

    key = sys.argv[1]
    config_path = get_config_path()

    if not config_path.exists():
        print(f"Error: Config file not found at {config_path}", file=sys.stderr)
        sys.exit(1)

    try:
        with open(config_path, 'r') as f:
            config = json.load(f)

        # Handle default_versions by deriving from supported_versions
        if key == "default_versions":
            if "supported_versions" not in config:
                print("Error: supported_versions not found in config", file=sys.stderr)
                sys.exit(1)
            result = list(config["supported_versions"].keys())
            print(",".join(result))
        elif key in config:
            result = config[key]
            if key == "supported_versions":
                # Output as JSON for shell parsing
                print(json.dumps(result))
            else:
                print(result)
        else:
            print(f"Error: Key '{key}' not found in config", file=sys.stderr)
            sys.exit(1)

    except Exception as e:
        print(f"Error reading config: {e}", file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()



import os
import unittest
from pathlib import Path
from .pytest_common import pushd, create_black_oil_simulator

class TestBasic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_dir = Path(os.path.dirname(__file__))
        cls.data_dir = test_dir.parent.joinpath("test_data/SPE1CASE1a")

    # IMPORTANT: Since all the python unittests run in the same process we must be
    #  careful to not call MPI_Init() more than once.
    #  Tests are run alphabetically, so we need to make sure that
    #  the the first test calls MPI_Init(), therefore the name of the tests
    #  have a numeric label like "01" in test_01_mpi_init to ensure that they
    #  are run in a given order.
    def test_01_mpi_init(self):
        with pushd(self.data_dir):
            sim = create_black_oil_simulator(filename="SPE1CASE1.DATA")
            sim.setup_mpi(init=True, finalize=False)
            sim.step_init()  # This will create the OPM::Main() object which will call MPI_Init()
            assert True

    def test_02_mpi_no_init(self):
        with pushd(self.data_dir):
            sim = create_black_oil_simulator(filename="SPE1CASE1.DATA")
            sim.setup_mpi(init=False, finalize=True)
            sim.step_init()  # This will create the OPM::Main() object which will not call MPI_Init()
            # That this test runs shows that the simulator does not call
            # MPI_Init() a second time
            assert True



import os
from contextlib import contextmanager

@contextmanager
def pushd(path):
    cwd = os.getcwd()
    if not os.path.isdir(path):
        os.makedirs(path)
    os.chdir(path)
    yield
    os.chdir(cwd)

ENABLE_ASYNC_ECL_OUTPUT_FLAG = '--enable-async-ecl-output=false'

def create_black_oil_simulator(*args, **kwargs):
    """Create BlackOilSimulator with test-safe default arguments.

    Automatically disables async ECL output to prevent race conditions
    with pushd context manager in tests.
    """
    from opm.simulators import BlackOilSimulator

    flag_to_add = ENABLE_ASYNC_ECL_OUTPUT_FLAG
    # Handle different constructor patterns
    if 'args' in kwargs:
        # Add our flag to existing args
        if flag_to_add not in kwargs['args']:
            kwargs['args'].append(flag_to_add)
    elif len(args) >= 4:
        # Constructor with deck, state, schedule, summary_config
        # Add args parameter
        if len(args) == 4:
            args = args + ([flag_to_add],)
        elif len(args) == 5:
            # Args already provided, add our flag
            args_list = list(args[4]) if args[4] else []
            if flag_to_add not in args_list:
                args_list.append(flag_to_add)
            args = args[:4] + (args_list,)
    else:
        # Constructor with filename - add args parameter
        kwargs['args'] = kwargs.get('args', [])
        if flag_to_add not in kwargs['args']:
            kwargs['args'].append(flag_to_add)

    return BlackOilSimulator(*args, **kwargs)

def create_gas_water_simulator(*args, **kwargs):
    """Create GasWaterSimulator with test-safe default arguments.

    Automatically disables async ECL output to prevent race conditions
    with pushd context manager in tests.
    """
    from opm.simulators import GasWaterSimulator

    flag_to_add = ENABLE_ASYNC_ECL_OUTPUT_FLAG
    # Add our flag to args
    kwargs['args'] = kwargs.get('args', [])
    if flag_to_add not in kwargs['args']:
        kwargs['args'].append(flag_to_add)

    return GasWaterSimulator(*args, **kwargs)

def create_onephase_simulator(*args, **kwargs):
    """Create OnePhaseSimulator with test-safe default arguments.

    Automatically disables async ECL output to prevent race conditions
    with pushd context manager in tests.
    """
    from opm.simulators import OnePhaseSimulator

    flag_to_add = ENABLE_ASYNC_ECL_OUTPUT_FLAG
    # Add our flag to args
    kwargs['args'] = kwargs.get('args', [])
    if flag_to_add not in kwargs['args']:
        kwargs['args'].append(flag_to_add)

    return OnePhaseSimulator(*args, **kwargs)



#!/usr/bin/env python3

# Generates a PDF with plots of all summary curves from a reference
# case and a 'new' simulation.

import argparse
from datetime import datetime, timedelta
from filelock import FileLock
import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
from opm.io.ecl import ESmry
import os
import pickle
from scipy import integrate, stats

# Run analysis of a test.
# Calculate the deviation for each curve
# and generate a pdf with plots ordered according to deviation
def run_analysis(ref_file_name, sys_file_name, test_name, ref_name, sim_name):
    ref_file = ESmry(ref_file_name + '.SMSPEC')
    sim_file = ESmry(sys_file_name + '.SMSPEC')

    ref_time = ref_file.dates()
    sim_time = sim_file.dates()

    ref_time_in_secs = [(v - ref_time[0]).total_seconds() for v in ref_time]
    sim_time_in_secs = [(v - sim_time[0]).total_seconds() for v in sim_time]

    plt.rcParams['font.size'] = 8

    # Attempt at sorting the graphs in descending eyeball norm order.
    # - Normalize by inf-norm to get the same range in each graph, ie (0,1).
    # - Convert graphs to probability distributions (ie integral under curve should be 1).
    # - Use the wasserstein distance scaled by area under reference curve.
    deviation={}
    for r in ref_file.keys():
        if r == 'TIME' or r == 'YEARS':
            continue
        try:
            ref = ref_file[r]
            sim = sim_file[r]
        except:
            continue

        if len(ref) == 0 and len(sim) == 0:
            continue

        if not (any(ref) or any(sim)):
            continue

        ref /= np.linalg.norm(ref, np.inf)
        sim /= np.linalg.norm(sim, np.inf)

        A_ref = integrate.trapezoid(ref, ref_time_in_secs, 0.0)
        A_sim = integrate.trapezoid(sim, sim_time_in_secs, 0.0)

        deviation[r] = stats.wasserstein_distance(ref / A_ref, sim / A_sim) * A_ref

    p = PdfPages(f'{test_name}.pdf')
    for r in sorted(deviation, key = lambda x: deviation[x], reverse=True):
        try:
            ref = ref_file[r]
            sim = sim_file[r]
        except:
            continue

        fig, ax = plt.subplots()
        ax.plot(ref_time, ref, linestyle='dashed', linewidth=0.5, marker='o', markersize=1.0)
        ax.plot(sim_time, sim, linewidth=0.5, marker='x', markersize=1.0)
        ax.legend([ref_name, sim_name])
        plt.title(r)
        u = ref_file.units(r)
        if u:
            plt.ylabel(u)
        myFmt = DateFormatter("%Y-%b")
        ax.xaxis.set_major_formatter(myFmt)
        ax.xaxis.set_major_locator(plt.MaxNLocator(20))
        plt.grid()
        fig.autofmt_xdate()
        fig.savefig(p, format='pdf')
        plt.close(fig)

    p.close()

    if os.path.exists('max_devs.pkl'):
        with FileLock('max_devs.lck'):
            with open('max_devs.pkl', 'rb') as f:
                max_deviations = pickle.load(f)
    else:
        max_deviations = {}

    max_dev = max(deviation, key = lambda x: deviation[x])
    max_deviations[test_name] = deviation[max_dev]

    with FileLock('max_devs.lck'):
        with open('max_devs.pkl', 'wb') as f:
            pickle.dump(max_deviations, f)

# Rename files to rank them according to maximum deviations
def reorder_files():
    with open('max_devs.pkl', 'rb') as f:
        max_deviations = pickle.load(f)

    c = 1
    for file in sorted(max_deviations, key = lambda x: max_deviations[x], reverse=True):
        os.rename(f'{file}.pdf', f'{c:02d}_{file}.pdf')
        c += 1

# Main code
parser = argparse.ArgumentParser('plot_well_comparison.py')

parser.add_argument('-c', help='Name of test to process', dest='test_name')
parser.add_argument('-r', help='Reference file', dest='ref_file')
parser.add_argument('-s', help='Simulation file', dest='sim_file')
parser.add_argument('-t', help='Name for first simulation', dest='ref_name', default='Reference')
parser.add_argument('-u', help='Name for second simulation', dest='sim_name', default='New simulation')
parser.add_argument('-o', choices=['plot', 'rename'], help='Operation to do', required=True, dest='operation')
args = parser.parse_args()

if args.operation == 'plot':
    print(f"Processing {args.test_name}")
    run_analysis(args.ref_file, args.sim_file, args.test_name, args.ref_name, args.sim_name)
else:
    reorder_files()



import os
import unittest
from pathlib import Path
from .pytest_common import pushd, create_black_oil_simulator

class TestBasic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_dir = Path(os.path.dirname(__file__))
        cls.data_dir = test_dir.parent.joinpath("test_data/SPE1CASE1a")

    def test_all(self):
        with pushd(self.data_dir):
            sim = create_black_oil_simulator(filename="SPE1CASE1.DATA")
            # NOTE: The following call should throw an exception since the simulation
            #   has not been initialized
            with self.assertRaises(RuntimeError):
                sim.get_dt()



import json
import os
import numpy as np
from pathlib import Path

from opm.io.ecl import ERst, ESmry

# Avoid tensorflow dependency but maintain compability with kerasify export_model method
class Dense:
    def __init__(self, input_dim, output_dim, activation='linear'):
        self.weights = np.zeros((input_dim, output_dim), dtype=np.float32)
        self.biases = np.zeros((output_dim,), dtype=np.float32)
        self.activation = activation

    def set_weights(self, w_b):
        self.weights = w_b[0].astype(np.float32)
        self.biases = w_b[1].astype(np.float32)

    def get_weights(self):
        return [self.weights, self.biases]

    def get_config(self):
        return {'activation': self.activation}

class Sequential:
    def __init__(self, layers):
        self.layers = layers


def extract_unrst_variables(data_dir, deck_file):
    """
    Extracts fluid state variables and timesteps from the .UNRST file
    corresponding to a deck.

    Parameters
    ----------
    data_dir : Path or str
        Directory containing the UNRST file.
    deck_file : str
        Name of the deck file (e.g., SPE1CASE1.DATA).

    Returns
    -------
    unrst_vars : dict
        Dictionary with keys "PRESSURE", "SWAT", "SOIL", "SGAS", "RS"
        Each value is a list of np.ndarray for each report step.
    dts : list
        List of timestep durations (approximated from report steps if needed).
    """
    deck_stem = Path(deck_file).stem
    unrst_file = str(Path(data_dir) / f"{deck_stem}.UNRST")  # convert to string
    unsmry_file = str(Path(data_dir) / f"{deck_stem}.SMSPEC")
    rst = ERst(unrst_file)
    unsmry = ESmry(unsmry_file)
    # Load all report steps
    report_steps = sorted(rst.report_steps)
    for step in report_steps:
        rst.load_report_step(step)

    variables_of_interest = ["PRESSURE", "SWAT", "SOIL", "SGAS", "RS"]
    unrst_vars = {var: [] for var in variables_of_interest}

    for step in report_steps:
        for var in variables_of_interest:
            if (var, step) in rst:
                unrst_vars[var].append(np.array(rst[var, step], copy=True))
            else:
                # Fill with zeros if variable not in UNRST for this step
                n_cells = len(rst["PRESSURE", step])  # assume same size as PRESSURE
                unrst_vars[var].append(np.zeros(n_cells, dtype=np.float32))

    # Compute approximate timesteps (differences between report steps)
    dates = unsmry.dates()
    times = [0.] + [(d - unsmry.start_date).total_seconds() for d in dates]

    return unrst_vars, times


def write_config(models, model_dir, multi_mode=False, case_label=None):
    """
    Write one or multiple model configs to JSON in OPM-supported format.
    """

    model_dir = Path(model_dir)
    if multi_mode and not case_label:
        raise ValueError("case_label must be provided for multi_mode")

    os.makedirs(model_dir, exist_ok=True)
    all_configs = {}

    for idx, m in enumerate(models):
        # Ensure input/output features are lists
        input_features = m["input_features"] if isinstance(m["input_features"], list) else [m["input_features"]]
        output_features = m["output_features"] if isinstance(m["output_features"], list) else [m["output_features"]]

        fe_input = m.get("feature_engineering_input") or [None]*len(input_features)
        fe_output = m.get("feature_engineering_output") or [None]*len(output_features)
        scaling_params_input = m.get("input_scaling_params") or [None]*len(input_features)
        scaling_params_output = m.get("output_scaling_params") or [None]*len(output_features)

        # Build input/output blocks
        input_block = {}
        for i, f in enumerate(input_features):
            d = {}
            if fe_input[i] and fe_input[i].lower() != "none":
                d["feature_engineering"] = fe_input[i].lower()
            if scaling_params_input[i] is not None:
                # convert numpy float32 to native float if needed
                d["scaling_params"] = {k: float(v) for k, v in scaling_params_input[i].items()}
            input_block[f] = d

        output_block = {}
        for i, f in enumerate(output_features):
            d = {}
            if fe_output[i] and fe_output[i].lower() != "none":
                d["feature_engineering"] = fe_output[i].lower()
            if scaling_params_output[i] is not None:
                d["scaling_params"] = {k: float(v) for k, v in scaling_params_output[i].items()}
            output_block[f] = d

        # Active cells
        active_cells = m.get("active_cells")
        if active_cells is None:
            n_cells = getattr(m, "n_cells", None)
            if n_cells is None:
                raise ValueError("Cannot determine active_cells; provide n_cells or active_cells")
            active_cells = list(range(n_cells))

        cells_file = model_dir / f"{os.path.splitext(os.path.basename(m['model_path']))[0]}_active_cells.txt"
        with open(cells_file, "w") as f:
            for cell in active_cells:
                f.write(f"{cell}\n")

        # Apply time(s)
        apply_time = m["apply_time"]
        if isinstance(apply_time, float):
            apply_times_dict = {"0": float(apply_time)}
        elif isinstance(apply_time, list):
            apply_times_dict = {str(i): float(t) for i, t in enumerate(apply_time)}
        else:
            raise ValueError("apply_time must be float or list of floats")

        model_name_rel = f"models/{os.path.basename(m['model_path'])}"
        cells_file_rel = f"models/{cells_file.name}"

        all_configs[str(idx)] = {
            "model_path": model_name_rel,
            "cell_indices_file": cells_file_rel,
            "apply_times": apply_times_dict,
            "features": {"inputs": input_block, "outputs": output_block}
        }

    # JSON filename
    if multi_mode:
        json_name = case_label + ".json"
    else:
        json_name = f"{os.path.splitext(os.path.basename(models[0]['model_path']))[0]}.json"

    json_path = model_dir / json_name
    with open(json_path, "w") as f:
        json.dump(all_configs, f, indent=2)

    return json_path


def seconds_to_days(sec):
    return sec / 86400.0


def apply_feature_engineering(data, method=None):
    if not method or method.lower() == "none":
        return data
    if method.lower() == "log":
        return np.log(data)
    if method.lower() == "log10":
        return np.log10(data)
    if method.lower() == "log1p":
        return np.log1p(data)
    raise ValueError(f"Unsupported feature engineering method: {method}")


def apply_scaling_with_params(data, method):
    data = np.array(data)
    if method == "standard":
        mean, std = np.mean(data), np.std(data)
        return ((data - mean) / std if std != 0 else data - mean), {"mean": mean, "std": std}
    if method == "minmax":
        mn, mx = np.min(data), np.max(data)
        return ((data - mn) / (mx - mn) if mx != mn else data - mn), {"min": mn, "max": mx}
    return data, None


def get_feature_data(name, timestep, unrst, rd_init, times=None, feat_eng=None):
    if name == "PERMX":
        val = np.array(rd_init["PERMX"])
    elif name == "SOIL":
        val = 1.0 - np.array(unrst["SGAS"][timestep]) - np.array(unrst["SWAT"][timestep])
    elif name == "TIMESTEP":
        if times is None:
            raise ValueError("times must be provided for TIMESTEP")
        val = times[timestep + 1] - times[timestep]
    else:
        val = np.array(unrst[name][timestep])
    return apply_feature_engineering(val, feat_eng)


def collect_input_features(input_features, start_idx, unrst, rd_init, times=None, feat_eng=None, scaling=None):
    input_data = []
    scaling_params = []

    for i, f in enumerate(input_features):
        fe = feat_eng[i] if feat_eng else None
        sc = scaling[i] if scaling else None
        val = get_feature_data(f, start_idx, unrst, rd_init, times=times, feat_eng=fe)
        val, params = apply_scaling_with_params(val, sc) if sc else (val, None)
        input_data.append(np.array(val).flatten())
        scaling_params.append(params)

    return np.concatenate(input_data), scaling_params


def compute_output_vars(output_features, start_idx, end_idx, unrst, rd_init, feature_engineering=None, scaling=None):
    output_data = []
    scaling_params = []

    for i, f in enumerate(output_features):
        fe = feature_engineering[i] if feature_engineering else None
        sc = scaling[i] if scaling else None

        if f.startswith("DELTA_"):
            base = f[6:]
            val = get_feature_data(base, end_idx, unrst, rd_init) - get_feature_data(base, start_idx, unrst, rd_init)
        else:
            val = get_feature_data(f, end_idx, unrst, rd_init)

        val = apply_feature_engineering(val, fe)
        val, params = apply_scaling_with_params(val, sc) if sc else (val, None)
        output_data.append(np.array(val).flatten())
        scaling_params.append(params)

    return np.stack(output_data), scaling_params


def extract_newtit_from_file(filename):
    """
    Parse an INFOSTEP file and return the values of the 'NewtIt' column.

    Parameters
    ----------
    filename : str or Path
        Path to the file (with or without .INFOSTEP extension).

    Returns
    -------
    list[int]
        List of NewtIt values for each data row.
    """
    path = Path(filename)
    if path.suffix.lower() != ".infostep":
        path = path.with_suffix(".INFOSTEP")
    if not path.exists():
        raise FileNotFoundError(f"INFOSTEP file not found: {path}")

    with path.open("r") as f:
        lines = [line.strip() for line in f if line.strip()]

    # Locate header with "NewtIt"
    header_idx = next((i for i, line in enumerate(lines) if "NewtIt" in line), None)
    if header_idx is None:
        raise ValueError(f"'NewtIt' column not found in {path}")

    header = lines[header_idx].split()
    idx = header.index("NewtIt")

    newt_values = []
    for row in lines[header_idx + 1 :]:
        cols = row.split()
        if len(cols) > idx:
            try:
                newt_values.append(int(cols[idx]))
            except ValueError:
                continue  # skip malformed rows

    return newt_values

# -------------------------
# Absolute (single timestep predictions)
# -------------------------
ABSOLUTE_CASES = [
    {
        "label": "single_RS",
        "mode": "single",
        "start_idx": 0, "end_idx": 1,
        "input_features": ["TIMESTEP"],
        "output_features": ["RS"],
    },
    {
        "label": "single_PRESSURE",
        "mode": "single",
        "start_idx": 0, "end_idx": 1,
        "input_features": ["PRESSURE"],
        "output_features": ["PRESSURE"],
    },
    {
        "label": "single_PERMX_to_SGAS",
        "mode": "single",
        "start_idx": 0, "end_idx": 1,
        "input_features": ["PERMX"],
        "output_features": ["SGAS"],
    },
    {
        "label": "single_SGAS_to_SWAT",
        "mode": "single",
        "start_idx": 0, "end_idx": 1,
        "input_features": ["SGAS"],
        "output_features": ["SWAT"],
    },
]

# -------------------------
# Relative / delta predictions
# -------------------------
RELATIVE_CASES = [
    {"label": "delta_RS", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["RS"], "output_features": ["DELTA_RS"]},
    {"label": "delta_SGAS", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["SGAS"], "output_features": ["DELTA_SGAS"]},
    {"label": "delta_SWAT", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["SWAT"], "output_features": ["DELTA_SWAT"]},
    {"label": "delta_PRESSURE", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PRESSURE"], "output_features": ["DELTA_PRESSURE"]},
    {"label": "delta_PERMX_to_RS", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PERMX"], "output_features": ["DELTA_RS"]},
    {"label": "delta_PERMX_to_PRESSURE", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PERMX"], "output_features": ["DELTA_PRESSURE"]},
]

# -------------------------
# Feature engineering cases
# -------------------------
FEATURE_ENGINEERING_CASES = [
    {"label": "feat_eng_RS_to_PRESSURE", "mode": "single", "start_idx": 0, "end_idx": 1,
     "input_features": ["RS"], "feature_engineering_input": ["log"],
     "output_features": ["PRESSURE"], "feature_engineering_output": ["log1p"]},
    {"label": "feat_eng_PRESSURE", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["SGAS"], "feature_engineering_input": ["log1p"],
     "output_features": ["PRESSURE"], "feature_engineering_output": ["log10"]},
    {"label": "feat_eng_PERMX_to_PRESSURE", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PERMX"], "feature_engineering_input": ["log"],
     "output_features": ["PRESSURE"], "feature_engineering_output": ["log"]},
]

# -------------------------
# Scaling cases
# -------------------------
SCALING_CASES = [
    {"label": "scaling_RS_minmax", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["RS"], "scaling_input": ["minmax"],
     "output_features": ["RS"], "scaling_output": ["minmax"]},
    {"label": "scaling_PRESSURE_standard", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PRESSURE"], "scaling_input": ["standard"],
     "output_features": ["PRESSURE"], "scaling_output": ["standard"]},
    {"label": "scaling_SGAS_SWAT_standard", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["SGAS", "SWAT"], "scaling_input": ["standard", "none"],
     "output_features": ["SGAS", "SWAT"], "scaling_output": ["none", "standard"]},
    {"label": "scaling_PRESSURE_SWAT_minmax", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PRESSURE", "SWAT"], "scaling_input": ["minmax", "none"],
     "output_features": ["SWAT", "PRESSURE"], "scaling_output": ["none", "minmax"]},
    {"label": "scaling_PERMX_TIMESTEP_standard", "mode": "single", "start_idx": 1, "end_idx": 2,
     "input_features": ["PRESSURE", "TIMESTEP"], "scaling_input": ["minmax", "none"],
     "output_features": ["SGAS", "SWAT"]},
]

# -------------------------
# Multi-model cases
# -------------------------
MULTI_MODEL_CASES = [
    {"label": "multi_same_index", "mode": "multi",
     "models": [
         {"input_features": ["PERMX"], "output_features": ["RS"], "start_idx": 1, "end_idx": 2},
         {"input_features": ["SGAS"], "output_features": ["SGAS"], "start_idx": 1, "end_idx": 2},
     ]},
    {"label": "multi_diff_index", "mode": "multi",
     "models": [
         {"input_features": ["PERMX"], "output_features": ["RS"], "start_idx": 1, "end_idx": 2},
         {"input_features": ["PRESSURE", "TIMESTEP", "PERMX"], "output_features": ["PRESSURE"], "start_idx": 8, "end_idx": 9},
         {"input_features": ["SGAS"], "output_features": ["SGAS"], "start_idx": 10, "end_idx": 11},
     ]},
]

ZERO_NEWTON_CASES = [
    {
        "label": "all_features",
        "mode": "multi",
        "models": [
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 0, "end_idx": 1},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 1, "end_idx": 2},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 2, "end_idx": 3},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 3, "end_idx": 4},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 4, "end_idx": 5},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 5, "end_idx": 6},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 6, "end_idx": 7},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 7, "end_idx": 8},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 8, "end_idx": 9},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 9, "end_idx": 10},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 10, "end_idx": 11},
            {"input_features": ["PERMX"], "output_features": ["RS", "PRESSURE", "SGAS", "SOIL"], "start_idx": 11, "end_idx": 12},
        ],

    },
]

ALL_CASES = [
    {
        "label": "all_cases",
        "mode": "multi",
        "models": [
            {"input_features": ["PERMX", "TIMESTEP", "RS", "SOIL", "SWAT", "PRESSURE"],
             "scaling_input": ["none", "none", "standard", "standard", "standard","minmax"],
             "feature_engineering_input": ["log", "log10", "log1p", "log1p", "log1p", "log10"],
             "output_features": ["RS", "PRESSURE", "DELTA_SGAS", "DELTA_SOIL"],
             "scaling_output": ["standard", "minmax", "standard", "standard"],
             "feature_engineering_output": ["log", "log10", "log1p", "log1p"],
             "start_idx": 0, "end_idx": 1},
             {"input_features": ["PERMX", "TIMESTEP", "RS", "SOIL", "SWAT", "PRESSURE"],
             "scaling_input": ["none", "none", "standard", "standard", "standard","minmax"],
             "feature_engineering_input": ["log", "log10", "log1p", "log1p", "log1p", "log10"],
             "output_features": ["DELTA_RS", "DELTA_PRESSURE", "DELTA_SGAS", "DELTA_SOIL"],
             "scaling_output": ["standard", "minmax", "standard", "standard"],
             "feature_engineering_output": ["none", "none", "log1p", "log1p"],
             "start_idx": 1, "end_idx": 2},
        ],

    },
]



import logging
import subprocess
import sys

from pathlib import Path

from opm_wheels.constants import Directories, PythonVersion


def canonicalize_docker_os(name: str) -> str:
    """Return a canonical filesystem/tag-friendly OS identifier.

    Accepts legacy values like "ubuntu:22.04" and returns "ubuntu-22.04".
    If already hyphenated (e.g., "ubuntu-22.04"), it is returned unchanged.
    """
    return name.replace(":", "-")

def get_docker_tag(docker_os: str) -> str:
    """Get the Docker tag for the given OS.

    The tag is always produced from the canonicalized OS name, e.g.:
    - input: "ubuntu:22.04" or "ubuntu-22.04" => tag: "test-opm-wheels-ubuntu-22.04"

    Supported values (canonical): ubuntu-22.04, ubuntu-24.04, debian-11.
    """
    canonical = canonicalize_docker_os(docker_os)
    return f"test-opm-wheels-{canonical}"

def get_git_root() -> Path:
    """Return the absolute path of the repository's root."""
    try:
        output = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"],
            stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError:
        # Handle the error if we're not inside a Git repo, etc.
        raise RuntimeError("Not a valid Git repository or other error occurred.")
    root = output.decode("utf-8").strip()
    # Extra check to ensure we're in the correct repository, check that there is a
    # directory called python/docker/test_wheels in the root directory

    if not Path(root).joinpath("python/docker/test_wheels").exists():
        raise RuntimeError("Current cwd is not inside repository.")
    return Path(root)

def get_testing_root_dir() -> Path:
    """Return the absolute path of the testing directory."""
    root = get_git_root()
    testing_root = root / "python" / "docker" / "test_wheels"
    if not testing_root.exists():
        raise RuntimeError("Testing directory does not exist.")
    return testing_root

def get_wheel_default_dir() -> Path:
    """Return the default wheel directory as a relative path."""
    wheel_default_dir = Path(Directories.python) / Directories.wheelhouse
    return wheel_default_dir

def get_wheel_abs_dir(wheel_dir: Path|str) -> Path:
    """Return the absolute path of the wheel directory."""
    # Check if the wheel_dir is an absolute path
    wheel_dir = Path(wheel_dir)
    if wheel_dir.is_absolute():
        return wheel_dir
    # If it's not absolute, resolve it relative to the python/docker directory
    git_root = get_git_root()
    docker_root = git_root / "python" / "docker"
    wheel_dir = docker_root / wheel_dir
    if not wheel_dir.exists():
        raise RuntimeError(f"Wheel directory {wheel_dir} does not exist.")
    return wheel_dir

def get_wheel_file(wheel_dir: str, name: str) -> Path:
    """Get the path to the wheel file. NAME is the name of a wheel file, relative to the
    python/wheelhouse directory of the Git repository."""
    wheel_dir = get_wheel_abs_dir(wheel_dir)
    wheel_file = wheel_dir / name
    if not wheel_file.exists():
        raise FileNotFoundError(f"Wheel file {wheel_file} does not exist")
    return wheel_file

def run_docker_build(
    docker_tag: str,
    opm_simulators_repo: str,
    opm_common_repo: str,
    opm_simulators_branch: str,
    opm_common_branch: str,
    build_ctxt_dir: Path,
    docker_file: Path,
    docker_progress: bool = False,
    python_versions: list[PythonVersion] = None,
) -> None:
    if docker_progress:
        progress="plain"
    else:
        progress="auto"
    # Build Docker command with base arguments
    docker_cmd = [
        "docker", "build",
        "--progress", progress,
        "--build-arg", "opm_simulators_repo=" + opm_simulators_repo,
        "--build-arg", "opm_common_repo=" + opm_common_repo,
        "--build-arg", "opm_simulators_branch=" + opm_simulators_branch,
        "--build-arg", "opm_common_branch=" + opm_common_branch,
    ]

    # Add Python versions build argument if specified
    if python_versions:
        python_versions_str = ",".join([str(v) for v in python_versions])
        docker_cmd.extend(["--build-arg", f"BUILD_PYTHON_VERSIONS={python_versions_str}"])
        logging.info(f"Building Docker image with Python versions: {python_versions_str}")

    # Add tag and context
    docker_cmd.extend([
        "-t", docker_tag,
        "-f", str(docker_file),
        str(build_ctxt_dir)
    ])

    try:
        result = subprocess.run(docker_cmd, check=True)
        logging.info(f"Docker image with tag {docker_tag} built successfully.")
    except subprocess.CalledProcessError as e:
        logging.error(f"Error building Docker image for tag {docker_tag}: {e}")
        sys.exit(1)
    return

def run_docker_run(
    docker_tag: str,
    wheel_dir: Path,
    python_versions: list[PythonVersion] | None,
    host_tests_dir: Path = None,
    test_cases_common: str = None,
    test_cases_simulators: str = None,
    stop_on_error: bool = False
) -> None:
    # Build Docker run command
    docker_cmd = [
        "docker", "run",
        "--rm",
        "-v", str(wheel_dir) + ":/test/wheels/",
    ]

    # Only set PYTHON_VERSIONS environment variable if explicitly specified
    # If not specified (None), let the container auto-detect installed versions
    if python_versions is not None:
        docker_cmd.extend(["-e", "PYTHON_VERSIONS=" + ",".join([str(v) for v in python_versions])])
        logging.info(f"Using specified Python versions: {','.join([str(v) for v in python_versions])}")
    else:
        logging.info("Using auto-detected Python versions from container")

    # Add host test directory mounts if specified
    if host_tests_dir:
        # Validate that the expected directories exist
        opm_common_python = host_tests_dir / "opm-common" / "python"
        opm_simulators_python = host_tests_dir / "opm-simulators" / "python"

        if not opm_common_python.exists():
            raise FileNotFoundError(f"Expected opm-common/python directory not found at: {opm_common_python}")
        if not opm_simulators_python.exists():
            raise FileNotFoundError(f"Expected opm-simulators/python directory not found at: {opm_simulators_python}")

        # Add volume mounts for test directories
        docker_cmd.extend([
            "-v", str(opm_common_python) + ":/test/opm-common/python",
            "-v", str(opm_simulators_python) + ":/test/opm-simulators/python"
        ])

        logging.info(f"Using host test directories from: {host_tests_dir}")

        # Signal to entrypoint that host directories are mounted
        docker_cmd.extend(["-e", "USE_HOST_TESTS=1"])

    # Add test selection environment variables if specified
    if test_cases_common:
        docker_cmd.extend(["-e", f"TEST_CASES_COMMON={test_cases_common}"])
        logging.info(f"Running only opm-common tests: {test_cases_common}")

    if test_cases_simulators:
        docker_cmd.extend(["-e", f"TEST_CASES_SIMULATORS={test_cases_simulators}"])
        logging.info(f"Running only opm-simulators tests: {test_cases_simulators}")

    # Add stop on error environment variable
    if stop_on_error:
        docker_cmd.extend(["-e", "STOP_ON_ERROR=1"])
        logging.info("Running with stop-on-error enabled")
    else:
        logging.info("Running with continue-on-error (default behavior)")

    # Add the Docker image tag
    docker_cmd.append(docker_tag)

    try:
        # Debug: log the full Docker command
        logging.info(f"DEBUG: Executing Docker command: {' '.join(docker_cmd)}")
        result = subprocess.run(
            docker_cmd,
            check=True,  # Raise an exception if the process returns a non-zero exit code
        )
        logging.info(f"Docker image with tag {docker_tag} run successfully.")
    except subprocess.CalledProcessError as e:
        logging.error(f"Error running Docker image for tag {docker_tag}: {e}")
        sys.exit(1)
    return
